"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import ABC, abstractmethod
from typing import Any


class FuseVisualizerBase(ABC):

    @abstractmethod
    def visualize(self, sample: Any, block: bool = True) -> None:
        """
        visualize sample
        :param sample: sample
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def visualize_aug(self, orig_sample: Any, aug_sample: Any, block: bool = True) -> None:
        """
        Visualise and compare augmented and non-augmented version of the sample
        :param orig_sample: original sample
        :param aug_sample: augmented sample
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        raise NotImplementedError
